import typing
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

from typeguard import typechecked

import tecton_core.tecton_pendulum as pendulum
from tecton_core.errors import TectonValidationError


if typing.TYPE_CHECKING:
    import pyspark


@dataclass
class MaterializationContext:
    _feature_start_time_DONT_ACCESS_DIRECTLY: Optional[datetime]
    _feature_end_time_DONT_ACCESS_DIRECTLY: Optional[datetime]
    _batch_schedule_DONT_ACCESS_DIRECTLY: timedelta
    _resources_DONT_ACCESS_DIRECTLY: Optional[Dict[str, Any]] = None
    _secrets_DONT_ACCESS_DIRECTLY: Optional[Dict[str, str]] = None

    @property
    def start_time(self) -> Optional[datetime]:
        return self._feature_start_time_DONT_ACCESS_DIRECTLY

    @property
    def end_time(self) -> Optional[datetime]:
        return self._feature_end_time_DONT_ACCESS_DIRECTLY

    @property
    def batch_schedule(self) -> timedelta:
        return self._batch_schedule_DONT_ACCESS_DIRECTLY

    @property
    def secrets(self) -> Optional[Dict[str, str]]:
        return self._secrets_DONT_ACCESS_DIRECTLY

    @property
    def resources(self) -> Optional[Dict[str, Any]]:
        return self._resources_DONT_ACCESS_DIRECTLY

    @typechecked
    def time_filter_sql(self, timestamp_expr: str) -> str:
        # Use atom string to include the timezone.
        return f"('{self.start_time.isoformat()}' <= ({timestamp_expr}) AND ({timestamp_expr}) < '{self.end_time.isoformat()}')"

    def time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        from pyspark.sql.functions import expr
        from pyspark.sql.functions import lit

        if isinstance(timestamp_expr, str):
            timestamp_expr = expr(timestamp_expr)

        # Use atom string to include the timezone.
        return (lit(self.start_time.isoformat()) <= timestamp_expr) & (timestamp_expr < lit(self.end_time.isoformat()))

    def feature_time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        return self.time_filter_pyspark(timestamp_expr)

    # Everything below is deprecated but kept for backwards compatibility since Snowflake Materialization Runtime
    # is not versioned.
    @property
    def feature_start_time(self) -> datetime:
        return self._feature_start_time_DONT_ACCESS_DIRECTLY

    @property
    def feature_end_time(self) -> datetime:
        return self._feature_end_time_DONT_ACCESS_DIRECTLY

    @property
    def feature_start_time_string(self) -> str:
        return self.feature_start_time.isoformat()

    @property
    def feature_end_time_string(self) -> str:
        return self.feature_end_time.isoformat()

    @typechecked
    def feature_time_filter_sql(self, timestamp_expr: str) -> str:
        return self.time_filter_sql(timestamp_expr)


@dataclass
class UnboundMaterializationContext(MaterializationContext):
    """
    This is only meant for instantiation in transformation default args. Using it directly will fail.
    """

    @property
    def batch_schedule(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def start_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def end_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def feature_start_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)

    @property
    def feature_end_time(self):
        msg = "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        raise TectonValidationError(msg)


@dataclass
class BoundMaterializationContext(MaterializationContext):
    @classmethod
    def create(
        cls,
        resources: Dict[str, Any],
        feature_start_time: Optional[pendulum.DateTime],
        feature_end_time: Optional[pendulum.DateTime],
        secrets: Optional[Dict[str, str]] = None,
    ) -> "BoundMaterializationContext":
        start_dt = None
        if feature_start_time:
            start_dt = datetime.fromtimestamp(feature_start_time.timestamp(), pendulum.tz.UTC)
        end_dt = None
        if feature_end_time:
            end_dt = datetime.fromtimestamp(feature_end_time.timestamp(), pendulum.tz.UTC)

        # user facing version
        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=start_dt,
            _feature_end_time_DONT_ACCESS_DIRECTLY=end_dt,
            # batch_schedule is passed by pipeline helper
            _batch_schedule_DONT_ACCESS_DIRECTLY=timedelta(seconds=0),
            _secrets_DONT_ACCESS_DIRECTLY=secrets,
            _resources_DONT_ACCESS_DIRECTLY=resources,
        )

    @classmethod
    def _create_internal(
        cls,
        feature_start_time: Optional[pendulum.DateTime],
        feature_end_time: Optional[pendulum.DateTime],
        batch_schedule: Optional[pendulum.Duration],
    ) -> "BoundMaterializationContext":
        start_dt = datetime.fromtimestamp(feature_start_time.timestamp(), pendulum.tz.UTC)
        end_dt = datetime.fromtimestamp(feature_end_time.timestamp(), pendulum.tz.UTC)

        # should only be used in pipeline_helper
        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=start_dt,
            _feature_end_time_DONT_ACCESS_DIRECTLY=end_dt,
            _batch_schedule_DONT_ACCESS_DIRECTLY=batch_schedule.as_timedelta(),
        )

    @classmethod
    def _create_from_period(
        cls, feature_time_limits: Optional[pendulum.Period], batch_schedule: pendulum.Duration
    ) -> "BoundMaterializationContext":
        feature_start_time = (
            feature_time_limits.start
            if feature_time_limits is not None
            else pendulum.from_timestamp(0, pendulum.tz.UTC)
        )
        feature_end_time = feature_time_limits.end if feature_time_limits is not None else pendulum.datetime(2100, 1, 1)

        start_dt = datetime.fromtimestamp(feature_start_time.timestamp(), pendulum.tz.UTC)
        end_dt = datetime.fromtimestamp(feature_end_time.timestamp(), pendulum.tz.UTC)

        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=start_dt,
            _feature_end_time_DONT_ACCESS_DIRECTLY=end_dt,
            _batch_schedule_DONT_ACCESS_DIRECTLY=batch_schedule.as_timedelta(),
        )


def materialization_context():
    """
    Used as a default value for a Feature View or Transformation with a materialization context parameter.

    ``context.start_time`` and ``context.end_time`` return a `datetime.datetime` object equal to the beginning and end of the period being materialized respectively. For example for a batch feature view materializing data from May 1st, 2022, ``context.start_time = datetime(2022, 5, 1)`` and ``context.end_time = datetime(2022, 5, 2)``.

    The datetimes can be used in SQL query strings directly (the datetime object will be cast to an atom-formatted timestamp string and inlined as a constant in the SQL query).

    Example usage:

    .. code-block:: python

        from tecton import batch_feature_view, materialization_context
        from datetime import datetime, timedelta

        @batch_feature_view(
            sources=[transactions],
            entities=[user],
            mode='spark_sql',
            features=[Attribute("AMOUNT", Float64)],
            batch_schedule=timedelta(days=1),
            feature_start_time=datetime(2020, 10, 10),
        )
        def user_last_transaction_amount(transactions, context=materialization_context()):
            return f'''
                SELECT
                    USER_ID,
                    AMOUNT,
                    TIMESTAMP
                FROM
                    {transactions}
                WHERE TIMESTAMP >= TO_TIMESTAMP("{context.start_time}") -- e.g. TO_TIMESTAMP("2022-05-01T00:00:00+00:00")
                    AND TIMESTAMP < TO_TIMESTAMP("{context.end_time}") -- e.g. TO_TIMESTAMP("2022-05-02T00:00:00+00:00")
                '''
    """
    dummy_time = pendulum.datetime(1970, 1, 1)
    dummy_period = timedelta()
    return UnboundMaterializationContext(
        _feature_start_time_DONT_ACCESS_DIRECTLY=dummy_time,
        _feature_end_time_DONT_ACCESS_DIRECTLY=dummy_time,
        _batch_schedule_DONT_ACCESS_DIRECTLY=dummy_period,
    )



from typing import Dict
from typing import List

from tecton._internals.errors import ExpressionParsingError
from tecton.types import SdkDataType
from tecton.vendor.sql_expresssion_parser import Lark_StandAlone
from tecton.vendor.sql_expresssion_parser import LexError
from tecton.vendor.sql_expresssion_parser import ParseError
from tecton.vendor.sql_expresssion_parser import Token
from tecton.vendor.sql_expresssion_parser import Transformer
from tecton.vendor.sql_expresssion_parser import VisitError
from tecton_core.data_types import BoolType
from tecton_core.data_types import Float32Type
from tecton_core.data_types import Float64Type
from tecton_core.data_types import Int32Type
from tecton_core.data_types import Int64Type
from tecton_core.data_types import StringType
from tecton_core.data_types import data_type_from_proto
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError
from tecton_proto.common.calculation_node__client_pb2 import AbstractSyntaxTreeNode
from tecton_proto.common.calculation_node__client_pb2 import DatePart
from tecton_proto.common.calculation_node__client_pb2 import LiteralValue
from tecton_proto.common.calculation_node__client_pb2 import Operation
from tecton_proto.common.calculation_node__client_pb2 import OperationType
from tecton_proto.common.data_type__client_pb2 import DataTypeEnum


class _ToAbstractSyntaxTree(Transformer):
    def __init__(self, original_expression: str, schema: Dict[str, SdkDataType]):
        self.original_expression = original_expression
        self.schema = schema
        super().__init__()

    def _type_is_numeric(self, node: AbstractSyntaxTreeNode) -> bool:
        return node.dtype in [Int64Type().proto, Int32Type().proto, Float64Type().proto, Float32Type().proto]

    def _types_are_comparable(self, nodes: List[AbstractSyntaxTreeNode]) -> bool:
        if not nodes:
            # let's assume empty lists are comparable
            return True
        elif all(self._type_is_numeric(node) for node in nodes):
            # All numeric types are comparable (might change this later)
            return True
        elif any(node.dtype.type == DataTypeEnum.DATA_TYPE_STRUCT for node in nodes):
            # structs are not comparable
            return False
        else:
            # if all types are the same
            first_node_type = nodes[0].dtype
            return all(node.dtype == first_node_type for node in nodes)

    def literal(self, value) -> AbstractSyntaxTreeNode:
        value = value[0]
        if isinstance(value, bool):
            literal_value = LiteralValue(bool_value=value)
            dtype = BoolType()
        elif isinstance(value, int):
            literal_value = LiteralValue(int64_value=value)
            dtype = Int64Type()
        elif isinstance(value, float):
            literal_value = LiteralValue(float64_value=value)
            dtype = Float64Type()
        elif isinstance(value, str):
            literal_value = LiteralValue(string_value=value)
            dtype = StringType()
        else:
            msg = f"Something went wrong while parsing: Unexpected Literal type {type(value)}, node: {value}"
            raise RuntimeError(msg)
        return AbstractSyntaxTreeNode(literal_value=literal_value, dtype=dtype.proto)

    def numeric_negative(self, child: List[Token]):
        return -1 * child[0]

    def INT(self, value: str):
        return int(value)

    def DECIMAL(self, value: str):
        return float(value)

    def FLOAT(self, value: str):
        return float(value)

    def ESCAPED_STRING(self, value: str):
        value = value[1:-1]  # strip parentheses
        return value

    def true(self, value: str):
        return True

    def false(self, value: str):
        return False

    def day(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.DAY)

    def month(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MONTH)

    def week(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.WEEK)

    def year(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.YEAR)

    def second(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.SECOND)

    def hour(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.HOUR)

    def minute(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MINUTE)

    def millennium(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MILLENNIUM)

    def century(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.CENTURY)

    def decade(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.DECADE)

    def quarter(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.QUARTER)

    def milliseconds(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MILLISECONDS)

    def microseconds(self, value: str):
        return AbstractSyntaxTreeNode(date_part=DatePart.MICROSECONDS)

    def COLUMN_REFERENCE(self, value: str):
        if value not in self.schema:
            msg = f"Invalid column reference: {value}, possible values: {list(self.schema.keys())}"
            raise ExpressionParsingError(msg)
        dtype = self.schema[value]
        return AbstractSyntaxTreeNode(column_reference=value, dtype=dtype.tecton_type.proto)

    def coalesce(self, values) -> AbstractSyntaxTreeNode:
        if not values:
            # This shouldn't be possible because the regex pattern for this rule requires at least 1 argument,
            # but adding as extra protection in case the rule changes in the future.
            msg = f"Error parsing expression: At least one value required for COALESCE, {self.original_expression}"
            raise ExpressionParsingError(msg)

        dtypes = {data_type_from_proto(value.dtype) for value in values}
        if len(dtypes) != 1:
            msg = f"Cannot mix values of types {dtypes} in COALESCE operator."
            raise ExpressionParsingError(msg)

        return AbstractSyntaxTreeNode(
            operation=Operation(operation=OperationType.COALESCE, operands=values), dtype=values[0].dtype
        )

    def datediff(self, values) -> AbstractSyntaxTreeNode:
        if not values or len(values) != 3:
            msg = f"Error parsing expression: Exactly three values required for DATEDIFF, {self.original_expression}"
            raise ExpressionParsingError(msg)

        date_part = values[0]
        start_date = values[1]
        end_date = values[2]
        if (
            start_date.dtype.type != DataTypeEnum.DATA_TYPE_TIMESTAMP
            or end_date.dtype.type != DataTypeEnum.DATA_TYPE_TIMESTAMP
        ):
            msg = "For DATEDIFF, both start_date and end_date must have type TIMESTAMP."
            raise ExpressionParsingError(msg)

        operands = [date_part, start_date, end_date]
        return AbstractSyntaxTreeNode(
            operation=Operation(operation=OperationType.DATE_DIFF, operands=operands), dtype=Int64Type().proto
        )

    def _comparison(self, values: List[AbstractSyntaxTreeNode], operation) -> AbstractSyntaxTreeNode:
        if len(values) != 2:
            # Shouldn't be possible due to regex patterns, so this is an internal error
            msg = f"Comparisons should have exactly 2 operands, found {len(values)}"
            raise TectonInternalError(msg)
        if not self._types_are_comparable(values):
            left_type = data_type_from_proto(values[0].dtype)
            right_type = data_type_from_proto(values[1].dtype)
            msg = f"Types {left_type} and {right_type} are not comparable"
            raise ExpressionParsingError(msg)
        return AbstractSyntaxTreeNode(operation=Operation(operation=operation, operands=values), dtype=BoolType().proto)

    def equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.EQUALS)

    def not_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.NOT_EQUALS)

    def greater_than(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.GREATER_THAN)

    def greater_than_or_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.GREATER_THAN_EQUALS)

    def less_than(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.LESS_THAN)

    def less_than_or_equals(self, values) -> AbstractSyntaxTreeNode:
        return self._comparison(values, OperationType.LESS_THAN_EQUALS)

    def _determine_math_dtype(self, values: List[AbstractSyntaxTreeNode], operation: OperationType):
        for value in values:
            if not self._type_is_numeric(value):
                msg = f"Cannot perform {operation} on non-numeric type {value.dtype}"
                raise ExpressionParsingError(msg)
        # TODO(FE-2517): do better datatype checking. This is just going to assume that everything is a Float64.
        return Float64Type

    def _math_operation(self, values, operation):
        dytpe = self._determine_math_dtype(values, operation)
        return AbstractSyntaxTreeNode(operation=Operation(operation=operation, operands=values), dtype=dytpe().proto)

    def addition(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.ADDITION)

    def subtraction(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.SUBTRACTION)

    def multiplication(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.MULTIPLICATION)

    def division(self, values) -> AbstractSyntaxTreeNode:
        return self._math_operation(values, OperationType.DIVISION)

    def negation(self, value):
        neg_1 = AbstractSyntaxTreeNode(literal_value=LiteralValue(int64_value=-1), dtype=Int32Type().proto)
        return self.multiplication([neg_1, value[0]])


def expression_to_proto(expression: str, schema: Dict[str, SdkDataType]) -> AbstractSyntaxTreeNode:
    # TODO: This class initialization should be top-level for better performance, but for not that causes
    #  issues with type-hint checking. We should fix that.
    lexer = Lark_StandAlone()
    try:
        lark_syntax_tree = lexer.parse(expression)
    except (LexError, ParseError) as e:
        msg = f"Error parsing expression: {expression}"
        raise ExpressionParsingError(msg) from e
    try:
        ast = _ToAbstractSyntaxTree(expression, schema).transform(lark_syntax_tree)
    except VisitError as e:
        if isinstance(e.orig_exc, TectonValidationError):
            raise e.orig_exc  # Re-raise the original TectonValidationError
        else:
            # This indicates an issue with our parsing logic, not a user error.
            msg = f"Error converting parse tree to AST: {lark_syntax_tree}"
            raise TectonInternalError(msg) from e
    return ast



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

CANARY_TYPE_BASE: CanaryType
CANARY_TYPE_NEW: CanaryType
CANARY_TYPE_UNSPECIFIED: CanaryType
DESCRIPTOR: _descriptor.FileDescriptor

class CanaryType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Python package for statistical functions in MLlib.
"""

from pyspark.mllib.stat._statistics import Statistics, MultivariateStatisticalSummary
from pyspark.mllib.stat.distribution import MultivariateGaussian
from pyspark.mllib.stat.test import ChiSqTestResult, KolmogorovSmirnovTestResult
from pyspark.mllib.stat.KernelDensity import KernelDensity

__all__ = ["Statistics", "MultivariateStatisticalSummary",
           "ChiSqTestResult", "KolmogorovSmirnovTestResult",
           "MultivariateGaussian", "KernelDensity"]



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1733347991', 'BUILD_TIMESTAMP': '1733347991', 'FORMATTED_DATE': '2024 Dec 04 21 33 11 Wed', 'GIT_COMMIT': '4378a6253fff9f73c1e816de53216e3804c20a70', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-medium-6b6f86cd5d-7lcgn', 'BUILD_USER': 'buildkite-agent'}
